(* Copyright (C) 2013 Benjamin Monmege 
   LSV, ENS Cachan & CNRS
   <benjamin.monmege@lsv.ens-cachan.fr>

   This program is free software: you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see
   <http://www.gnu.org/licenses/>. *)

module Matrix =
struct
  let map f m =
    Array.mapi (fun i -> Array.mapi (fun j -> f i j)) m
  let eye n =
    Array.init n (fun i -> Array.init n (fun j -> if i=j then 1. else 0.))
  exception Out_Of_Bound
  (* copy the submatrix of m1 of size x,y starting from coefficient i1,j1 *)
  (* into the submatrix of m2 starting from coefficient i2,j2 *)
  let blit m1 i1 j1 m2 i2 j2 x y =
    if Array.length m1 < i1 + x ||
      Array.length m2 < i2 + x ||
      Array.length m1.(0) < j1 + y ||
      Array.length m2.(0) < j2 + y 
    then raise Out_Of_Bound
    else
      for i=0 to x-1 do 
	for j=0 to y-1 do
	  m2.(i2+i).(j2+j) <- m1.(i1+i).(j1+j)
	done
      done
  let recopy m1 m2 =
    blit m1 0 0 m2 0 0 (Array.length m1) (Array.length m1.(0))
  let submatrix m i j x y =
    let m' = Array.make_matrix x y m.(0).(0) in
    blit m i j m' 0 0 x y;
    m'
  let copy m = submatrix m 0 0 (Array.length m) (Array.length m.(0))
  let right_external_product m s mult =
    Array.map (fun x -> mult x s) m
  let left_external_product s m mult =
    Array.map (fun x -> mult s x) m
  let add m1 m2 plus =
    map (fun i -> fun j -> plus m1.(i).(j)) m2
  exception Product_Impossible
  let prod m1 m2 plus mult zero = 
    let n =  Array.length m1 
    and p = Array.length m2 
    and q =  Array.length m2.(0) in
    if Array.length m1.(0) <> p then raise Product_Impossible
    else
      let m = Array.make_matrix n q zero in
      for i=0 to n-1 do
	for j=0 to q-1 do
	  for k=0 to p-1 do
	    m.(i).(j) <- plus m.(i).(j) (mult m1.(i).(k) m2.(k).(j))
	  done
	done
      done;
      m
  let tensorial_product m1 m2 mult =
    map (fun i -> fun j -> mult m1.(i).(j)) m2
  let vectorial_product v1 v2 mult =
    Array.init (Array.length v1) 
      (fun i -> Array.init (Array.length v2) (fun j -> mult v1.(i) v2.(j)))
  exception Star_Impossible
  let rec star m1 plus mult zero = 
    let n = Array.length m1 in
    if Array.length m1.(0) <> n then raise Star_Impossible
    else
      let m = Array.make_matrix n n m1.(0).(0) in
      if n = 1 then 
	begin
	  if m1.(0).(0) = 1. then raise Star_Impossible 
	  else m.(0).(0) <- 1./.(1.-.m1.(0).(0))
	end
      else
	begin
	  let a = submatrix m1 0 0 (n-1) (n-1) 
	  and b = submatrix m1 0 (n-1) (n-1) 1
	  and c = submatrix m1 (n-1) 0 1 (n-1)
	  and d = submatrix m1 (n-1) (n-1) 1 1 in
	  let d' = star d plus mult zero in
	  let a' = star (add a (prod b (prod d' c plus mult zero) 
				  plus mult zero) plus) plus mult zero in
	  let b' = prod a' (prod b d' plus mult zero) plus mult zero in
	  let c' = prod d' (prod c a' plus mult zero) plus mult zero in
	  let d'' = add d' (prod d' 
			      (prod c (prod a' (prod b d' plus mult zero)
					 plus mult zero) plus mult zero)
			      plus mult zero) plus in
	  blit a' 0 0 m 0 0 (n-1) (n-1);
	  blit b' 0 0 m 0 (n-1) (n-1) 1;
	  blit c' 0 0 m (n-1) 0 1 (n-1);
	  blit d'' 0 0 m (n-1) (n-1) 1 1;
	end;
      m
end;;

module Word =
struct
  type letter = |First of char |Last |Middle of char
  let find_letter u i =
    if i=0 then First u.[0]
    else if i=String.length u then Last
    else Middle u.[i]
end;;

module Formula = 
struct 
  include Word
  type form = 
  |True |Begin |End |Letter of char |Var of char |Not of form |And of form*form |Or of form*form 
  let rec simplify = function
    |True -> True
    |Begin -> Begin
    |End -> End
    |Letter a -> Letter a
    |Var x -> Var x
    |Not f1 -> 
      let f1' = simplify f1 in
      begin
	match f1' with
	|Not g1 -> g1
	|And (g1,g2) -> simplify (Or (Not g1,Not g2))
	|Or (g1,g2) -> simplify (And (Not g1,Not g2)) 
	|_ -> Not f1'
      end
    |And (f1,f2) -> 
      let f1' = simplify f1 and f2' = simplify f2 in
      begin
	match f1',f2' with
	|True,_ -> f2'
	|_,True -> f1'
	|_,_ -> And (f1',f2')
      end
    |Or (f1,f2) -> 
      let f1' = simplify f1 and f2' = simplify f2 in
      begin
	match f1',f2' with
	|True,_ -> f2'
	|_,True -> f1'
	|_,_ -> And (f1',f2')
      end
  let rec gastex_print = function
    |True -> "\\top"
    |Letter a -> "\\basictest{" ^ Char.escaped a ^ "}"
    |Begin -> "\\inittest"
    |End -> "\\finaltest"
    |Var x -> "\\basictest{" ^ Char.escaped x ^ "}"
    |Not f' -> "\\lnot " ^ (gastex_print f')
    |And (f1,f2) -> (gastex_print f1) ^ " \\land " ^ (gastex_print f2)
    |Or (f1,f2) -> (gastex_print f1) ^ " \\lor " ^ (gastex_print f2)
  let rec print = function
    |True -> "True"
    |Letter a -> "Letter " ^ Char.escaped a
    |Begin -> "Begin"
    |End -> "End"
    |Var x -> "Var " ^ Char.escaped x
    |Not f' -> "Not (" ^ (print f') ^ ")"
    |And (f1,f2) -> "And (" ^ (print f1) ^ "," ^ (print f2) ^ ")"
    |Or (f1,f2) -> "Or (" ^ (print f1) ^ "," ^ (print f2) ^ ")"
  let mult f1 f2 = simplify (And(f1,f2))
  (*exception Not_found_precise of char * int*)
  let rec models u sigma i = function 
    |True -> i<String.length u
    |Begin -> i=0
    |End -> i=String.length u -1
    |Letter a -> i<String.length u && u.[i]=a
    |Var x -> (try List.assoc x sigma = i with Not_found -> false)
    |Not f' -> not (models u sigma i f')
    |And (f1,f2) -> (models u sigma i f1) && (models u sigma i f2)
    |Or (f1,f2) -> (models u sigma i f1) || (models u sigma i f2)
end;;

module Move = 
struct		
  type move = |RightM |LeftM |Drop of char |Lift |Final 
  (* Final only used for final states for uniformity reasons *)
  type term= Formula.form * move
  exception Empty
  let mult (f1,m1) (f2,m2) = 
    match m1,m2 with
    |Final,Final -> (Formula.mult f1 f2,Final)
    |Final,_ -> (Formula.mult f1 f2,m2)
    |_,Final -> (Formula.mult f1 f2,m1)
    |_,_ when m1=m2 -> (Formula.mult f1 f2,m1)
    |_,_ -> raise Empty
  let gastex_print (f,m) =
    (Formula.gastex_print f) ^ 
      (match m with
      |RightM -> "\\dright"
      |LeftM -> "\\dleft"
      |Drop x -> "\\drop x"
      |Lift -> "\\lift"
      |Final -> "")
end;;

module Polynomial =
struct
  type term = Move.term
  type polynomial = (float * term) list
  let zero = []
  let monom s x = [(s,x)]
  let rec add p = function 
    |[] -> p
    |(s,x) :: p' -> (s,x) :: (add p p')
  let rec prod_monom (s,x) = function
    |[] -> []
    |(s',y) :: p -> 
      try let z = Move.mult x y in (s *. s',z) :: (prod_monom (s,x) p) with
	Move.Empty -> prod_monom (s,x) p
  let rec prod p = function
    |[] -> []
    |(s,x) :: p' -> add (prod_monom (s,x) p) (prod p p')
  let rec star = function
    |[] -> [1.,(Formula.True,Move.Final)]
    |(s,x) :: p' -> (1. /. (1. -. s),x) :: (star p')
  let rec gastex_print = function
    |[] -> ""
    |[(s,x)] -> (string_of_float s) ^ " " ^ (Move.gastex_print x) ^ ""
    |(s,x) :: p -> (string_of_float s) ^ " " ^ (Move.gastex_print x) ^ " + " ^ (gastex_print p)
  let rec eval u sigma i m = function 
    |[] -> 0.
    |(s,(f,m')) :: p -> if m = m' && Formula.models u sigma i f 
      then s +. (eval u sigma i m p)
      else eval u sigma i m p
  let rec pebbles = function
    |[] -> []
    |(s,(f,Move.Drop x)) :: p -> x :: (pebbles p)
    |_ :: p -> pebbles p
  let rec extract_transition = function
    |[] -> []
    |(s,(f,Move.RightM)) :: p -> (s,(f,Move.RightM))::(extract_transition p)
    |(s,(f,Move.LeftM)) :: p -> (s,(f,Move.LeftM))::(extract_transition p)
    |_ :: p -> extract_transition p
  let rec extract_drop = function
    |[] -> []
    |(s,(f,Move.Drop x)) :: p -> (s,(f,Move.Drop x))::(extract_drop p)
    |_ :: p -> extract_drop p
  let rec extract_lift = function
    |[] -> []
    |(s,(f,Move.Lift)) :: p -> (s,(f,Move.Lift))::(extract_lift p)
    |_ :: p -> extract_lift p
(*
  let rec extract_final = function
    |[] -> []
    |(s,(f,Move.Final)) :: p -> (s,(f,Move.Final))::(extract_final p)
    |_ :: p -> extract_final p
*)
end;;


module LayeredAutomaton =
struct
  type layer =
    {nbstates: int;  
     (* if states = n, then the actual states of the layer are
	{0,1,2,...,n-1} *)
     transition: Polynomial.polynomial array array; 
     (* transition matrix is predecomposed into internal transitions
	of the layer *)
     drop: Polynomial.polynomial array array; 
     (* and drop/lift transitions *)
     lift: Polynomial.polynomial array array}
  type automaton = 
    {layers: layer list;
     initial: float array;
     final: Polynomial.polynomial array}
  (* initialization of a standard layered automaton with n states and
     k layers *)
  let init n n' = 
    (* initialize a standard one-layer automaton with n states, and
       virtual n' states in next layer *)
    {layers = [{nbstates = n;
		transition = Array.make_matrix n n Polynomial.zero;
		drop = Array.make_matrix n n' Polynomial.zero;
		lift = Array.make_matrix n' n Polynomial.zero}];
     initial = Array.init n (fun i -> if i=0 then 1. else 0.);
     final = Array.make n Polynomial.zero}
  let add_layer auto n = 
    (* plug lauto into a new automaton with n more states in fresh
       layer 0 *)
    let n' = (List.hd auto.layers).nbstates in
    {layers =
	{nbstates = n;
	 transition = Array.make_matrix n n Polynomial.zero;
	 drop = Array.make_matrix n n' Polynomial.zero;
	 lift = Array.make_matrix n' n Polynomial.zero} :: auto.layers;
     initial = Array.init n (fun i -> if i=0 then 1. else 0.);
     final = Array.make n Polynomial.zero}
  let number_of_states lays =
    (* number of states of the first layer *)
    match lays with
    |[] -> failwith "no layers"
    |lay::_ -> lay.nbstates
  let eval_transition lay u sigma i m =
    Matrix.map (fun p -> fun q -> Polynomial.eval u sigma i m) lay.transition 
  let eval_drop lay u sigma i m =
    Matrix.map (fun p -> fun q -> Polynomial.eval u sigma i m) lay.drop 
  let eval_lift lay u sigma i =
    Matrix.map (fun p -> fun q -> Polynomial.eval u sigma i Move.Lift) lay.lift 
  let eval_nested layer0 u sigma i peb matrix_layer1 =
    let nest = Array.make_matrix layer0.nbstates layer0.nbstates 0. in
    for x = 0 to Array.length peb - 1 do 
      Matrix.recopy 
	(Matrix.add nest 
	   (Matrix.prod (eval_drop layer0 u sigma i (Move.Drop peb.(x))) 
	      (Matrix.prod (matrix_layer1.(x).(i)) 
		 (eval_lift layer0 u ((peb.(x),i)::sigma) 
		    (String.length u-1)) 
		 ( +. ) ( *. ) 0.) 
	      ( +. ) ( *. ) 0.) 
	   ( +. )) nest;
    done;
    nest
  let rec eval_matrix lays u sigma peb =
    (* lays: layer array *)
    (* u: string *)
    (* sigma: (char * int) list *)
    (* peb: char array *)
    match lays with
    |[] -> failwith "attention empty automaton"
    |[lay] -> 
      let leftloop = Matrix.eye lay.nbstates in
      let right = Matrix.copy leftloop in
      for i = 1 to String.length u-1 do
	let temp_right = eval_transition lay u sigma (i-1) Move.RightM in
	Matrix.recopy 
	  (Matrix.star 
	     (Matrix.prod (eval_transition lay u sigma i Move.LeftM) 
		(Matrix.prod leftloop temp_right ( +. ) ( *. ) 0.) 
		( +. ) ( *. ) 0.) 
	     ( +. ) ( *. ) 0.) leftloop;
	Matrix.recopy 
	  (Matrix.prod right 
	     (Matrix.prod temp_right leftloop ( +. ) ( *. ) 0.) 
	     ( +. ) ( *. ) 0.) right
      done;
      right
    |lay::lays' -> 
      let matrix_layer1 = Array.make_matrix (Array.length peb) (String.length u) 
	(Array.make_matrix (number_of_states lays') (number_of_states lays') 0.) in
      for x = 0 to Array.length peb - 1 do
	for i = 0 to String.length u - 1 do
	  matrix_layer1.(x).(i) <- eval_matrix lays' u ((peb.(x),i)::sigma) peb;
	done
      done;
      let leftloop = Matrix.star (eval_nested lay u sigma 0 peb matrix_layer1) ( +. ) ( *. ) 0. in
      let right = Matrix.copy leftloop in
      for i = 1 to String.length u - 1 do
	let temp_right = eval_transition lay u sigma (i-1) Move.RightM in
	Matrix.recopy 
	  (Matrix.star
	     (Matrix.add (eval_nested lay u sigma i peb matrix_layer1)
		(Matrix.prod (eval_transition lay u sigma i Move.LeftM) 
		   (Matrix.prod leftloop temp_right ( +. ) ( *. ) 0.) 
		   ( +. ) ( *. ) 0.) 
		( +. ))
	     ( +. ) ( *. ) 0.) leftloop;
	Matrix.recopy (Matrix.prod right 
			 (Matrix.prod temp_right leftloop ( +. ) ( *. ) 0.) 
			 ( +. ) ( *. ) 0.) right
      done;
      right
  let rec union l = function
    |[] -> l
    |x :: l' -> if List.mem x l then union l l' else x :: (union l l')
  let rec pebbles lays =
    match lays with
    |[] -> []
    |lay::lays' ->
      Array.fold_left 
	(fun l -> Array.fold_left 
	  (fun l -> fun p -> union l (Polynomial.pebbles p)) l) 
	(pebbles lays') lay.drop
  let eval auto u =
    let mat = eval_matrix auto.layers u [] (Array.of_list (pebbles auto.layers)) 
    and n = number_of_states (auto.layers) in 
    let result = ref 0. in 
    for p = 0 to n - 1 do
      for q = 0 to n - 1 do
	result := !result +. 
	  auto.initial.(p) *. mat.(p).(q) *. 
	  (Polynomial.eval u [] (String.length u-1) Move.Final auto.final.(q))
      done;
    done;
    !result
end;;

module Automaton = (* Automata in extension: better for compositional Kleene theorem *)
struct
  type automaton =
    {nbstates: int;  (* if states = n, then the actual states are {0,1,2,...,n-1} *)
     initial: float array;
     final: Polynomial.polynomial array;
     transition: Polynomial.polynomial array array}
  type layeredAutomaton = 
    {auto: automaton;
     nblayers: int;
     layers: int array} (* array of size nbstates, that associates to each state its layer *)
  (* initialization of a standard layered automaton with n states and k layers *)
  let init n k =
    {auto =
	{nbstates = n;
	 initial = Array.init n (fun i -> if i=0 then 1. else 0.);
	 final = Array.make n Polynomial.zero;
	 transition = Array.make_matrix n n Polynomial.zero};
     nblayers = k;
     layers = Array.make n 0}
  let states_in_layer lauto k = (* 0 <= k <= lauto.nblayers - 1 *)
    let rec aux i =
      if i = lauto.auto.nbstates then (0,[])
      else
	begin
	  let (m,l) = aux (i+1) in
	  if lauto.layers.(i) = k then (m+1,i::l) else (m,l)
	end
    in let (m,l) = aux 0 in (m,Array.of_list l)
  let extract_transition auto m mapping =
    let tab = Array.make_matrix m m Polynomial.zero in
    for p=0 to m-1 do
      for q=0 to m-1 do
	tab.(p).(q) <- Polynomial.extract_transition auto.transition.(mapping.(p)).(mapping.(q))
      done
    done;
    tab
  let extract_drop auto m mapping m' mapping' =
    let tab = Array.make_matrix m m' Polynomial.zero in
    for p=0 to m-1 do
      for q=0 to m'-1 do
	tab.(p).(q) <- Polynomial.extract_drop auto.transition.(mapping.(p)).(mapping'.(q))
      done
    done;
    tab
  let extract_lift auto m' mapping' m mapping =
    let tab = Array.make_matrix m' m Polynomial.zero in
    for p=0 to m'-1 do
      for q=0 to m-1 do
	tab.(p).(q) <- Polynomial.extract_lift auto.transition.(mapping'.(p)).(mapping.(q))
      done
    done;
    tab
  let extract_tab tab m mapping =
    let tab' = Array.make m tab.(0) in
    for p=0 to m-1 do
      tab'.(p) <- tab.(mapping.(p))
    done;
    tab'
  let rec extract_layers lauto k = 
    if k = lauto.nblayers then [] 
    else if k = lauto.nblayers - 1 then
      begin
	let (m,mapping) = states_in_layer lauto k in
	{LayeredAutomaton.nbstates = m;
	 LayeredAutomaton.transition = extract_transition lauto.auto m mapping;
	 LayeredAutomaton.drop = Array.make m [||];
	 LayeredAutomaton.lift = [||]} :: (extract_layers lauto (k+1))
      end
    else 
      begin
	let (m,mapping) = states_in_layer lauto k 
	and (m',mapping') = states_in_layer lauto (k+1) in
	{LayeredAutomaton.nbstates = m;
	 LayeredAutomaton.transition = extract_transition lauto.auto m mapping;
	 LayeredAutomaton.drop = extract_drop lauto.auto m mapping m' mapping';
	 LayeredAutomaton.lift = extract_lift lauto.auto m' mapping' m mapping} 
	:: (extract_layers lauto (k+1))
      end
  let translation lauto = (* translate in previous formalism *)
    let (m,mapping) = states_in_layer lauto 0 in
    {LayeredAutomaton.layers = extract_layers lauto 0;
     LayeredAutomaton.initial = extract_tab lauto.auto.initial m mapping;
     LayeredAutomaton.final = extract_tab lauto.auto.final m mapping}
  let eval lauto u =
    LayeredAutomaton.eval (translation lauto) u
  let print lauto = 
    ""
  let gastex_print lauto hsep vsep =
    let output = ref "\\begin{gpicture}\n" in
    output := !output ^ "\t\\gasset{fangle=-90,loopdiam=4,Nw=4,Nh=4}\n";
    for i=0 to lauto.auto.nbstates-1 do
      output := !output ^ "\t\\node(" ^ (string_of_int i) ^ ")(" 
      ^ (string_of_float (hsep*.(float_of_int i))) ^ "," ^ 
	(string_of_float (-.vsep*.(float_of_int (lauto.layers.(i)-1)))) ^ 
	"){$" ^ (string_of_int i) ^ "$}\n";
      if lauto.auto.initial.(i)<>0. then
	output := !output ^ "\t\\imark(" ^ (string_of_int i) ^ 
	  ")\n\t\\node[Nframe=n]()(" ^ (string_of_float (hsep*.((float_of_int i)-. 0.4))) ^ 
	  "," ^ (string_of_float (-.vsep*.(float_of_int (lauto.layers.(i)-1))+.1.)) ^ "){$" ^ 
	  (string_of_float lauto.auto.initial.(i)) ^ "$}\n";
      if lauto.auto.final.(i)<>Polynomial.zero then
	output := !output ^ "\t\\fmark(" ^ (string_of_int i) ^ ")\n\t\\node[Nframe=n]()(" ^ 
	  (string_of_float (hsep*.((float_of_int i)+.0.3))) ^ "," ^ 
	  (string_of_float (-.vsep*.(float_of_int (lauto.layers.(i)-1))-.vsep*.0.4)) ^ "){$" ^ 
	  (Polynomial.gastex_print lauto.auto.final.(i)) ^ "$}\n";
    done;
    for i=0 to lauto.auto.nbstates-1 do
      for j=0 to lauto.auto.nbstates-1 do
	if lauto.auto.transition.(i).(j)<>Polynomial.zero then
	  if i<>j then
	    output := !output ^ "\t\\drawedge(" ^ (string_of_int i) ^ "," ^ 
	      (string_of_int j) ^ "){$" ^ 
	      (Polynomial.gastex_print lauto.auto.transition.(i).(j)) ^ "$}\n"
	  else
	     output := !output ^ "\t\\drawloop(" ^ (string_of_int i) ^ "){$" ^ 
	      (Polynomial.gastex_print lauto.auto.transition.(i).(j)) ^ "$}\n"
      done
    done;
    !output ^ "\\end{gpicture}\n"
end;;


module Expression = 
struct
  type expression =
  |Val of float |Form of Formula.form |Right |Left 
  |Plus of expression*expression |Conc of expression*expression
  |Star of expression |PStar of expression |Compute of char*expression
  (*let rec print_expression = function 
    |Val s -> string_of_float s
    |Form f -> Formula.print f
    |Right -> "\\dright"
    |Left -> "\\dleft"
    |Plus (e1,e2) -> (print_expression e1) ^ " + " ^ (print_expression e2)
    |Conc (e1,e2) -> (print_expression e1) ^ "\\conc " ^ (print_expression e2)
    |Star e1 -> "(" ^ (print_expression e1) ^ ")^*"
    |PStar e1 -> "(" ^ (print_expression e1) ^ ")^+"
    |Compute (x,e1) -> "(\\compute{" ^ Char.escaped x ^ "}{" ^ (print_expression e1) ^ "})"*)
  let rec print_expression = function 
    |Val s -> "Val " ^ (string_of_float s)
    |Form f -> "Form (" ^ Formula.print f ^ ")"
    |Right -> "Right"
    |Left -> "Left"
    |Plus (e1,e2) -> "Plus (" ^ (print_expression e1) ^ "," ^ (print_expression e2) ^ ")"
    |Conc (e1,e2) -> "Conc (" ^ (print_expression e1) ^ "," ^ (print_expression e2) ^ ")"
    |Star e1 -> "Star (" ^ (print_expression e1) ^ ")"
    |PStar e1 -> "PStar (" ^ (print_expression e1) ^ ")"
    |Compute (x,e1) -> "Compute (" ^ Char.escaped x ^ "," ^ (print_expression e1) ^ ")"
end;;


module Kleene = 
struct
  open Automaton
  open Expression
  (* FROM EXPRESSIONS TO AUTOMATA *)
  (* construct the standard automaton with unique initial state 0 *)
  let rec automaton_of_expression = function	
    |Val s -> 
      let lauto = init 1 1 in 
      lauto.auto.final.(0) <- Polynomial.monom s (Formula.True,Move.Final);
      lauto
    |Form f ->
      let lauto = init 1 1 in 
      lauto.auto.final.(0) <- Polynomial.monom 1. (f,Move.Final);
      lauto
    |Right -> 
      let lauto = init 2 1 in 
      lauto.auto.final.(1) <- Polynomial.monom 1. (Formula.True,Move.Final);
      lauto.auto.transition.(0).(1) <- Polynomial.monom 1. (Formula.True,Move.RightM);
      lauto
    |Left -> 
      let lauto = init 2 1 in 
      lauto.auto.final.(1) <- Polynomial.monom 1. (Formula.True,Move.Final);
      lauto.auto.transition.(0).(1) <- Polynomial.monom 1. (Formula.True,Move.LeftM);
      lauto
    |Plus (e1,e2) -> 
      let lauto1 = automaton_of_expression e1 and lauto2 = automaton_of_expression e2 in
      let n1 = lauto1.auto.nbstates and n2 = lauto2.auto.nbstates in
      let lauto = init (n1 + n2 - 1) (max lauto1.nblayers lauto2.nblayers) in
      lauto.auto.final.(0) <- Polynomial.add lauto1.auto.final.(0) lauto2.auto.final.(0);
      Array.blit lauto1.auto.final 1 lauto.auto.final 1 (n1-1);
      Array.blit lauto2.auto.final 1 lauto.auto.final n1 (n2-1);
      Array.blit lauto1.auto.transition.(0) 1 lauto.auto.transition.(0) 1 (n1-1);
      Array.blit lauto2.auto.transition.(0) 1 lauto.auto.transition.(0) n1 (n2-1);
      Matrix.blit lauto1.auto.transition 1 1 lauto.auto.transition 1 1 (n1-1) (n1-1);
      Matrix.blit lauto2.auto.transition 1 1 lauto.auto.transition n1 n1 (n2-1) (n2-1);
      Array.blit lauto1.layers 1 lauto.layers 1 (n1-1);
      Array.blit lauto2.layers 1 lauto.layers n1 (n2-1);
      lauto
    |Conc (e1,e2) -> 
      let lauto1 = automaton_of_expression e1 and lauto2 = automaton_of_expression e2 in
      let n1 = lauto1.auto.nbstates and n2 = lauto2.auto.nbstates in
      let lauto = init (n1 + n2 - 1) (max lauto1.nblayers lauto2.nblayers) in
      lauto.auto.final.(0) <- Polynomial.prod lauto1.auto.final.(0) lauto2.auto.final.(0);
      Array.blit
	(Matrix.right_external_product lauto1.auto.final lauto2.auto.final.(0) Polynomial.prod)
	1 lauto.auto.final 1 (n1-1);
      Array.blit lauto2.auto.final 1 lauto.auto.final n1 (n2-1);
      Array.blit lauto1.auto.transition.(0) 1 lauto.auto.transition.(0) 1 (n1-1);
      Array.blit 
	(Matrix.left_external_product lauto1.auto.final.(0) lauto2.auto.transition.(0) 
	   Polynomial.prod)
	1 lauto.auto.transition.(0) n1 (n2-1);
      Matrix.blit lauto1.auto.transition 1 1 lauto.auto.transition 1 1 (n1-1) (n1-1);
      for i=1 to n1-1 do 
	for j=n1 to n1+n2-2 do 
	  lauto.auto.transition.(i).(j) <- Polynomial.prod lauto1.auto.final.(i) 
	    lauto2.auto.transition.(0).(j-n1+1)
	done;
      done;
      Matrix.blit lauto2.auto.transition 1 1 lauto.auto.transition n1 n1 (n2-1) (n2-1);
      Array.blit lauto1.layers 1 lauto.layers 1 (n1-1);
      Array.blit lauto2.layers 1 lauto.layers n1 (n2-1);
      lauto
    |Star e1 -> 
      let lauto1 = automaton_of_expression e1 in
      let n = lauto1.auto.nbstates in
      let cstar = Polynomial.star lauto1.auto.final.(0) in
      let lauto = init n lauto1.nblayers in 
      lauto.auto.final.(0) <- cstar;
      Array.blit (Matrix.left_external_product cstar lauto1.auto.final Polynomial.prod) 1 
	lauto.auto.final 1 (n-1);
      Array.blit (Matrix.left_external_product cstar lauto1.auto.transition.(0) Polynomial.prod) 1
	lauto.auto.transition.(0) 1 (n-1);
      Matrix.blit 
	(Matrix.add lauto1.auto.transition 
	   (Matrix.vectorial_product 
	      (Matrix.left_external_product cstar lauto1.auto.final Polynomial.prod) 
	      lauto1.auto.transition.(0) Polynomial.prod)
	   Polynomial.add) 1 1
	lauto.auto.transition 1 1 (n-1) (n-1);
      Array.blit lauto1.layers 1 lauto.layers 1 (n-1);
      lauto
    |PStar e1 ->
      let lauto1 = automaton_of_expression e1 in
      let n = lauto1.auto.nbstates in
      let cstar = Polynomial.star lauto1.auto.final.(0) in
      let lauto = init n lauto1.nblayers in 
      lauto.auto.final.(0) <- Polynomial.prod cstar lauto1.auto.final.(0);
      Array.blit (Matrix.left_external_product cstar lauto1.auto.final Polynomial.prod) 1 
	lauto.auto.final 1 (n-1);
      Array.blit (Matrix.left_external_product cstar lauto1.auto.transition.(0) Polynomial.prod) 1
	lauto.auto.transition.(0) 1 (n-1);
      Matrix.blit 
	(Matrix.add lauto1.auto.transition 
	   (Matrix.vectorial_product 
	      (Matrix.left_external_product cstar lauto1.auto.final Polynomial.prod) 
	      lauto1.auto.transition.(0) Polynomial.prod)
	   Polynomial.add) 1 1
	lauto.auto.transition 1 1 (n-1) (n-1);
      Array.blit lauto1.layers 1 lauto.layers 1 (n-1);
      lauto
    |Compute (var,e1) ->
      let lauto1 = automaton_of_expression e1 in
      let n = lauto1.auto.nbstates in
      let lauto = init (n+2) (lauto1.nblayers + 1) in
      lauto.auto.final.(n+1) <- Polynomial.monom 1. (Formula.True,Move.Final);
      lauto.auto.transition.(0).(1) <- Polynomial.monom 1. (Formula.True,Move.Drop var);
      Matrix.blit lauto1.auto.transition 0 0 lauto.auto.transition 1 1 n n;
      for i=1 to n do
	lauto.auto.transition.(i).(n+1) <- 
	  Polynomial.prod lauto1.auto.final.(i-1) (Polynomial.monom 1. (Formula.End,Move.Lift))
      done;
      for i=1 to n do
	lauto.layers.(i) <- lauto1.layers.(i-1) + 1
      done;
      lauto
  let eval_expression e u = eval (automaton_of_expression e) u
end;;
